# pylint: disable=no-member, invalid-name
# pylint: disable=too-many-arguments, too-many-instance-attributes
"""
Load .et4 file into mem (experimental).
This file structure may be modified in near future.
"""
# Copyright (c) Benyuan Liu. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
from struct import unpack

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt


class ET4:
    """.et4 file loader"""

    def __init__(
        self, file_name, ex_mtx=None, step=1, compatible=False, output_resistor=False
    ):
        """
        initialize .et4 handler.
        .et4 is an experimental file format for XEIT-ng system

        try to read data and parse FILE HEADER
            [-- 128 uint parameters + (256 float RE) + (256 float IM) --]

        Parameters
        ----------
        file_name : basestring
        ex_mtx : NDArray
            num_lines x 2 array
        step : int
            architecture of voltage meter
        compatible : bool
            if data output needs to be .et3 compatible
        output_resistor : bool
            convert voltage to current

        Returns
        -------
        NDArray
            return data (complex valued)

        Notes
        -----
        in .et4, a frame consists (16) excitations, and in each excitation
        ADC samples are arranged as AD1, AD2, ..., AD16, consecutively. so
        (AD1, AD2, ..., AD16) (AD1, AD2, ..., AD16) ... (AD1, AD2, ..., AD16)

        "see hardware section in eit.pdf, benyuan liu."

        when the excitation electrodes are marked as A, B,
        in order to be compatible with .et3, you should
        1. automatically zero-out 4 measures where A, A-1, B, B-1
        2. rearrange all the measures so that in each excitation,
           the first measurement always start from (include) 'A'
        3. divide 'diff Voltage' / Current = Resistor (optional)
        """
        self.file_name = file_name
        self.ex_mtx = ex_mtx
        self.step = step
        self.compatible = compatible
        self.output_resistor = output_resistor

        # 1. get .et4 file length
        nbytes = et4_tell(self.file_name)

        # 2. get nframes (a frame = (256 + 256d + 256d) = 5120 Bytes)
        self.info_num = 256
        self.data_num = 256 * 2
        self.header_size = self.info_num * 4
        self.frame_size = self.header_size + self.data_num * 8
        self.nframe = int((nbytes) / (self.frame_size))

        # 3. load data
        self.data = self.load()

    def load(self):
        """load RAW data"""
        # 1. prepare storage
        x = np.zeros((self.nframe, self.data_num), dtype=np.double)

        # 3. unpack data and extract parameters
        with open(self.file_name, "rb") as fh:
            for i in range(self.nframe):
                d = fh.read(self.frame_size)
                x[i] = np.array(unpack("512d", d[self.header_size :]))

        data = x[:, :256] + 1j * x[:, 256:]
        # electrode re-arranged the same as .et3 file
        if self.compatible:
            v_index, c_index = zero_rearrange_index(self.ex_mtx)
            dout = data[:, v_index]
            if self.output_resistor:
                # number of diff_V per stimulation
                M = int(len(v_index) / len(c_index))
                # current index is the same in each M measurements
                cm_index = np.repeat(c_index, M)
                # R = diff_V / I
                dout = dout / data[:, cm_index]
        else:
            dout = data

        return dout

    def load_info(self):
        """load info headers from xEIT"""
        # 1. prepare storage
        info = np.zeros((self.nframe, 256))

        # 3. unpack data and extract parameters
        with open(self.file_name, "rb") as fh:
            for i in range(self.nframe):
                d = fh.read(self.frame_size)
                info[i, :] = np.array(unpack("33if222i", d[: self.header_size]))

        return info

    def to_df(self, resample=None, rel_date=None, fps=20):
        """convert raw data to pandas.DataFrame"""
        if rel_date is None:
            rel_date = "2019/01/01"
        ta = np.arange(self.nframe) * 1.0 / fps
        ts = pd.to_datetime(rel_date) + pd.to_timedelta(ta, "s")
        df = pd.DataFrame(self.data, index=ts)
        # resample
        if resample is not None:
            df = df.resample(resample).mean()

        return df

    def to_csv(self):
        """save file to csv"""
        raise NotImplementedError()


def et4_tell(fstr):
    """check the filetype of et4"""
    with open(fstr, "rb") as fh:
        fh.seek(0, 2)  # move the cursor to the end (2) of the file
        file_len = fh.tell()

    return file_len


def zero_rearrange_index(ex_mtx):
    """
    (default mode: opposition stimulation)
    0. excitation electrodes are denoted by 'A' and 'B'
    1. for each excitation, REARRANGE all the data start from 'A'
    2. zero all the channels of A, A-1, B, B-1

    returns : re-ordered non-zero index, current index
    """
    if ex_mtx is None:
        num_lines, num_el, el_dist = 16, 16, 8
        ab_scan = False
    else:
        num_lines, num_el = ex_mtx.shape
        ab_scan = True

    v_index, c_index = [], []  # non-zero diff-pairs and current values
    for k in range(num_lines):
        if ab_scan:
            ex_pat = ex_mtx[k, :].ravel()
            a = np.where(ex_pat == 1)[0][0]
            b = np.where(ex_pat == -1)[0][0]
        else:
            a = k  # positive excitation
            b = (a + el_dist) % num_el  # negative excitation
        ap = (a - 1) % num_el  # positive adjacent
        bp = (b - 1) % num_el  # negative adjacent
        # print(A, B, Ap, Bp)
        c_index.append(k * num_el + b)
        for i in range(num_el):
            # re-order data start after A
            j = (i + a) % num_el
            if j not in (a, b, ap, bp):
                v_index.append(k * num_el + j)

    return v_index, c_index


if __name__ == "__main__":

    # .et4 file
    et_file = r"C:\xeit\eit_20190929-103342.et4"
    # et_file = "../../datasets/s00-02.et4"

    # load data
    et4 = ET4(et_file, compatible=True, output_resistor=False)
    et4_data = et4.data
    print(et4_data.shape)

    ti = et4_data.sum(axis=1) / 192.0
    ti_real = np.real(ti)
    ti_imag = np.imag(ti)
    ti_abs = np.sqrt(ti_real**2 + ti_imag**2)
    print("max = ", np.max(ti_abs))
    print("min = ", np.min(ti_abs))

    xlim = 1000
    if ti_abs.shape[0] < 1000:
        xlim = ti_abs.shape[0]

    # plot
    fig = plt.figure(figsize=(6, 4))
    ax = fig.add_subplot(211)
    ax.plot(ti_real, "b-")
    axt = ax.twinx()
    axt.plot(ti_imag, "r-")
    ax.set_xlim([0, xlim])
    ax.grid(True)

    ax2 = fig.add_subplot(212)
    ax2.plot(ti_abs, "r-")
    ax2.grid(True)
    ax2.set_xlim([0, xlim])
    plt.show()
    # fig.savefig('s00-03-80k.png')



# pylint: disable=no-member, invalid-name
# pylint: disable=too-many-arguments, too-many-locals
"""
Dynamic EIT imaging and information retrieval from EIT images
"""
# Copyright (c) Benyuan Liu. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
from pyeit.eit import bp
from pyeit.eit import jac


class DynamicEIT:
    """dynamic eit imaging"""

    def __init__(
        self, mesh=None, el_pos=None, parser="fmmu", algo="jac", p=0.20, lamb=0.001
    ):
        """initialize"""
        if algo == "jac":
            solver = jac.JAC(mesh, el_pos, perm=1.0, parser=parser)
            solver.setup(p=p, lamb=lamb, method="kotre")
        else:
            # default: 'bp'
            solver = bp.BP(mesh, el_pos, parser="fmmu", step=1)
            solver.setup(weight="simple")

        self.solver = solver

    def normalize(self, v1, v0):
        """normalize according to ref frame"""
        raise NotImplementedError

    def map(self, v):
        """map boundary voltages to EIT images"""
        raise NotImplementedError



""" add customized plot for 2D/3D mesh """
from .voronoi_plot import voronoi_plot

try:
    from .tetplot import tetplot
except ImportError:
    print("mesh.plot: vispy is required for 3D plotting")

__all__ = ["voronoi_plot", "tetplot"]



# coding: utf-8
# pylint: disable=invalid-name, too-many-arguments
# pylint: disable=too-many-instance-attributes
"""
This is a python code template that guide you through
writing your own reconstruction algorithms.
"""
# Copyright (c) Benyuan Liu. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
from __future__ import division, absolute_import, print_function

import numpy as np

from .fem import Forward
from .utils import eit_scan_lines


class EitBase:
    """
    A base EIT solver.
    """

    def __init__(
        self,
        mesh,
        el_pos,
        ex_mat=None,
        step=1,
        perm=None,
        jac_normalized=False,
        parser="std",
    ):
        """
        Parameters
        ----------
        mesh: dict
            mesh structure
        el_pos: array_like
            position (numbering) of electrodes
        ex_mat: array_like, optional (default: opposition)
            2D array, each row is one stimulation pattern/line
        step: int, optional
            measurement method
        perm: array_like, optional
            initial permittivity in generating Jacobian
        jac_normalized: Boolean (default is False)
            normalize the jacobian using f0 computed from input perm
        parser: str, optional, default is 'std'
            parsing the format of each frame in measurement/file

        Notes
        -----
        parser is required for your code to be compatible with
        (a) simulation data set or (b) FMMU data set
        """
        if ex_mat is None:
            ex_mat = eit_scan_lines(len(el_pos), 8)
        if perm is None:
            perm = mesh["perm"]

        # build forward solver
        fwd = Forward(mesh, el_pos)
        self.fwd = fwd

        # solving mesh structure
        self.mesh = mesh
        self.pts = mesh["node"]
        self.tri = mesh["element"]

        # shape of the mesh
        self.no_num, self.n_dim = self.pts.shape
        self.el_num, self.n_vertices = self.tri.shape
        self.el_pos = el_pos
        self.parser = parser

        # user may specify a scalar for uniform permittivity
        if np.size(perm) == 1:
            self.perm = perm * np.ones(self.el_num)
        else:
            self.perm = perm

        # solving configurations
        self.ex_mat = ex_mat
        self.step = step

        # solving Jacobian using uniform sigma distribution
        res = fwd.solve_eit(ex_mat, step=step, perm=self.perm, parser=self.parser)
        self.J, self.v0, self.B = res.jac, res.v, res.b_matrix

        # Jacobian normalization: divide each row of J (J[i]) by abs(v0[i])
        if jac_normalized:
            self.J = self.J / np.abs(self.v0[:, None])

        # initialize other parameters
        self.params = {}
        self.xg = []
        self.yg = []
        self.mask = []

        # mapping matrix
        self.H = self.B.T
        # warning: user must run solver.setup() manually to get correct H
        # self.setup()

    def setup(self):
        """setup EIT solver"""
        raise NotImplementedError

    def solve(self, v1, v0, normalize=False, log_scale=False):
        """
        dynamic imaging (conductivities imaging)

        Parameters
        ----------
        v1: NDArray
            current frame
        v0: NDArray
            referenced frame, d = H(v1 - v0)
        normalize: Bool, optional
            true for conducting normalization
        log_scale: Bool, optional
            remap reconstructions in log scale

        Returns
        -------
        ds: NDArray
            complex-valued NDArray, changes of conductivities
        """
        if normalize:
            dv = self.normalize(v1, v0)
        else:
            dv = v1 - v0

        ds = -np.dot(self.H, dv.transpose())  # s = -Hv
        if log_scale:
            ds = np.exp(ds) - 1.0

        return ds

    def map(self):
        """simple mat using projection matrix"""
        raise NotImplementedError

    def normalize(self, v1, v0):
        """
        Normalize current frame using the amplitude of the reference frame.
        Boundary measurements v are complex-valued, we can use the real part of v,
        np.real(v), or the absolute values of v, np.abs(v).

        Parameters
        ----------
        v1: NDArray
            current frame, can be a Nx192 matrix where N is the number of frames
        v0: NDArray
            referenced frame, which is a row vector
        """
        dv = (v1 - v0) / np.abs(v0)

        return dv



"""
pyEIT: Python based Electrical Impedance Tomography

eit:
    reconstruction algorithms
mesh:
    create triangle(2D) / tetrahedron(3D) meshes
"""
__author__ = "Benyuan Liu"
__author_email__ = "liubenyuan@gmail.com"
__license__ = "BSD"
APP_NAME = "pyeit"
APP_VERSION = "1.1.8"
APP_DESCRIPTION = "A Python Package for EIT"



# coding: utf-8
# pylint: disable=invalid-name, no-member
""" implement quality assessment functions for distmesh """
# Copyright (c) Benyuan Liu. All rights reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
from __future__ import division, absolute_import, print_function


def stats(p, t):
    """
    print mesh or tetrahedral status

    Parameters
    ----------
    p: array_like
        coordinates of nodes (x, y) in 2D, (x, y, z) in 3D
    t: array_like
        connectives forming elements

    Notes
    -----
    a simple function for illustration purpose only.
    print the status (size) of nodes and elements
    """
    print("mesh status:")
    print("%d nodes, %d elements" % (p.shape[0], t.shape[0]))



# pylint: disable=no-member, invalid-name, too-many-locals
"""
load daeger .eit files
reimplement daeger-eit in EIDORS3D (eidors_readdata.m)

You need to create a ex_mat using eit_scan_lines with 16 electrodes and
adjacent stimulation, rotate measurements in eit simulation.
"""
# Copyright (c) Benyuan Liu. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
import os
import struct
import numpy as np


class DAEGER_EIT:
    """process daeger pulmovista .eit file"""

    def __init__(self, fname):
        """
        initialize

        Note
        ----
        Daeger using standard measuremants, basic_stim in EIDORS, where
        mk_stim_patterns(16, 1, [0, 1], [0, 1], ... \
                         {'no_meas_current','rotate_meas'}, 0.005);
        1, 16 electrodes, 1 ring,
        2, adjacent stimulation, adjacent measurement, [ex_mat, step=1]
        3, skip current carring electrodes,
        4, rotate measurements with stimulation patterns
        5, 1 unit current

        Parameters
        ----------
        file_name : basestring
            file path
        """
        self.fname = fname
        self.info = self.read_header(self.fname)
        self.ft = np.array([0.00098242, 0.00019607])  # estimated AA: 2016-04-07

    @staticmethod
    def read_header(fname, max_lines=50):
        """read information from header in text format"""
        fr = 0
        fmt = 0
        with open(fname, "r", encoding="ISO-8859-1") as fd:
            data = fd.readlines()
            lines = data[:max_lines]
            for line in lines:
                if "Framerate [Hz]" in line:
                    cell = line.split(":")
                    fr = int(cell[1])
                if "Format:" in line:
                    cell = line.split(":")
                    fmt = int(cell[1])
        # failed to find valid parameters
        if fr == 0:
            print("Frame rate could not be read, setting to 20")
            fr = 20
        if fmt == 0:
            print("Format could not be read, setting to 51")
            fmt = 51

        # find spc: bytes per frame
        daeger_spc = dict({31: 4112, 32: 3200, 51: 5495})
        if fmt not in daeger_spc.keys():
            print("Error, format version={} not supported".format(fmt))
        spc = daeger_spc[fmt]

        # get file length
        flen = os.path.getsize(fname)

        # get data offset
        with open(fname, "rb") as fh:
            b = fh.read(16)
            a = struct.unpack("8H", b)
            offset = a[2] + 8

        # number of frames
        nframe = int((flen - offset) / spc)

        # information in headers
        par = {
            "framerate": fr,
            "format": fmt,
            "spc": spc,
            "flen": flen,
            "offset": offset,
            "nframe": nframe,
        }

        return par

    def read_data(self):
        """read data frame by frame"""
        nframe = self.info["nframe"]
        data = np.zeros((nframe, 600), dtype=np.double)
        with open(self.fname, "rb") as fh:
            fh.seek(self.info["offset"])
            for i in range(nframe):
                d = fh.read(self.info["spc"])
                data[i] = struct.unpack("600d", d[:4800])

        return data

    def load(self):
        """convert data in to measurements (voltages)"""
        data = self.read_data()
        vv = self.ft[0] * data[:, :208] - self.ft[1] * data[:, 322:530]
        return vv

    def to_df(self):
        """
        Require a data structure document within a frame of daeger pulmovista,
        to extract the time stamp information and build the time series index.
        """
        raise NotImplementedError()


if __name__ == "__main__":
    file_name = r"./ID_SC_10_001.eit"
    model = DAEGER_EIT(fname=file_name)
    print(model.info)
    # test loading from daeger EIT
    model.load()



# coding: utf-8
# pylint: disable=no-member, invalid-name, too-many-arguments
# pylint: disable=unused-argument, unused-variable, function-redefined
""" plot function based on vispy.visuals for tetrahedral plots """
from __future__ import absolute_import

# vispy
from vispy.visuals import CompoundVisual
from vispy.visuals.mesh import MeshVisual
from vispy.visuals.line import LineVisual
from vispy.visuals.markers import MarkersVisual
from vispy.color import Color


# build your visual
class TetVisual(CompoundVisual):
    """display a 3D mesh"""

    def __init__(
        self,
        vertices=None,
        simplices=None,
        vertex_colors=None,
        edge_color=None,
        edge_width=1,
        markers=None,
        marker_colors=None,
        marker_size=1,
        **kwargs
    ):
        """
        a mesh visualization toolkit that can also plot edges or markers

        Parameters
        ----------

        Notes
        -----
        """
        self._mesh = MeshVisual()
        self._edge = LineVisual()
        self._edge_color = Color(edge_color)
        self._marker = MarkersVisual()
        #
        self._vertex_colors = vertex_colors

        self._update()
        # initialize visuals
        CompoundVisual.__init__(self, [self._mesh, self._edge, self._marker], **kwargs)
        # set default state, 'opaque', 'translucent' or 'additive'
        self._mesh.set_gl_state(
            preset="translucent",
            blend=True,
            depth_test=False,
            cull_face=False,
            polygon_offset_fill=True,
            polygon_offset=(1, 1),
        )
        # end
        self.freeze()

        def _update(self):
            """
            update parameters to visuals
            """
            raise NotImplementedError()

        @property
        def edge_color(self):
            """get"""
            return self._edge_color

        @edge_color.setter
        def edge_color(self, edge_color):
            """set"""
            self._edge_color = Color(edge_color)
            self._update()



# coding: utf-8
# pylint: disable=invalid-name, no-member, too-many-instance-attributes
# pylint: disable=too-many-arguments, arguments-differ
"""
GREIT (using distribution method)

Note, that, the advantages of greit is NOT on simulated data, but
1. construct RM using real-life data with a stick move in the cylinder
2. construct RM on finer mesh, and use coarse-to-fine map for visualization
3. more robust to noise by adding noise via (JJ^T + lamb*Sigma_N)^{-1}
"""
# Copyright (c) Benyuan Liu. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
from __future__ import division, absolute_import, print_function

import numpy as np
import scipy.linalg as la

from .base import EitBase
from .interp2d import meshgrid, weight_sigmod


class GREIT(EitBase):
    """The GREIT algorithm"""

    def setup(self, method="dist", w=None, p=0.20, lamb=1e-2, n=32, s=20.0, ratio=0.1):
        """
        setup GREIT

        Parameters
        ----------
        method: str, optional
            'set' or 'dist'
        w: NDArray, optional
            weight on each element
        p: float, optional
            noise covariance
        lamb: float
            regularization parameters
        n: int, optional
            grid size
        s: float, optional
            control the blur
        ratio: float, optional
            desired ratio

        References
        ----------
        [1] Bartlomiej Grychtol, Beat Muller, Andy Adler
            "3D EIT image reconstruction with GREIT"
        [2] Adler, Andy, et al.
            "GREIT: a unified approach to 2D linear EIT reconstruction of
            lung images." Physiological measurement 30.6 (2009): S35.
        """
        # parameters for GREIT projection
        if w is None:
            w = np.ones_like(self.mesh["perm"])
        self.params = {"w": w, "p": p, "lamb": lamb, "n": n, "s": s, "ratio": ratio}
        # action (currently only support 'dist')
        if method == "dist":
            w_mat, self.xg, self.yg, self.mask = self._build_grid()
            self.H = self._build_dist(w_mat)
        else:
            raise ValueError("method " + method + " not supported yet")

    def map(self, dv):
        """return H*v"""
        return -np.dot(self.H, dv.transpose())

    def _build_dist(self, w_mat):
        """generate R using distribution method."""
        lamb, p = self.params["lamb"], self.params["p"]

        f = self.fwd.solve_eit(
            self.ex_mat, step=self.step, perm=self.perm, parser=self.parser
        )
        jac = f.jac
        # E[yy^T], it is more efficient to use left pinv than right pinv
        j_j_w = np.dot(jac, jac.T)
        r_mat = np.diag(np.diag(j_j_w) ** p)
        jac_inv = la.inv(j_j_w + lamb * r_mat)
        # RM = E[xx^T] / E[yy^T]
        h_mat = np.dot(np.dot(w_mat.T, jac.T), jac_inv)

        return h_mat

    def _build_grid(self):
        """build grids and mask"""
        # initialize grids
        n = self.params["n"]
        xg, yg, mask = meshgrid(self.pts, n=n)
        # mapping from values on triangles to values on grids
        xy = np.mean(self.pts[self.tri], axis=1)
        xyi = np.vstack((xg.flatten(), yg.flatten())).T
        # GREIT is using sigmod as weighting function (global)
        ratio, s = self.params["ratio"], self.params["s"]
        w_mat = weight_sigmod(xy, xyi, ratio=ratio, s=s)
        return w_mat, xg, yg, mask

    def get_grid(self):
        """get grids and mask"""
        return self.xg, self.yg, self.mask

    def mask_value(self, ds, mask_value=0):
        """(plot only) mask values on nodes outside 2D mesh."""
        ds[self.mask] = mask_value
        ds = ds.reshape(self.xg.shape)
        return self.xg, self.yg, ds

    @staticmethod
    def build_set(x, y):
        """generate R from a set of training sets (deprecate)."""
        # E_w[yy^T]
        y_y_t = la.inv(np.dot(y, y.transpose()))
        h_matrix = np.dot(np.dot(x, y), y_y_t)
        return h_matrix



# coding: utf-8
# pylint: disable=invalid-name, no-member, too-many-locals
""" create multi-layered mesh on a unit circle """
# Copyright (c) Benyuan Liu. All rights reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
from __future__ import division, absolute_import, print_function

import numpy as np
import matplotlib.pyplot as plt


class MeshCircle:
    """create meshes on uniform circle"""

    def __init__(self, n_fan=6, n_layer=8, n_el=16):
        """
        Parameters
        ----------
        n_fan : int
            number of fans (see the inner most layer)
        n_layer : int
            number of layers
        n_el : int
            number of boundary electrodes (default: 16)
        """
        self.n_fan = n_fan
        self.n_layer = n_layer
        self.n_el = n_el

        # number of points per-layer
        pts_per_layer = self.n_fan * np.arange(self.n_layer + 1)
        # The number of points of the initial layer (center point: 1 node)
        pts_per_layer[0] = 1
        self.pts_per_layer = pts_per_layer

        # starting point index of each layer
        # the initial layer (put at [-1]) must start with 0
        index = np.cumsum(pts_per_layer)
        index[-1] = 0
        self.index_per_layer = index

    def create(self):
        """create pts and tri"""
        pts = self._spawn_points()
        tri = self._spawn_elements()
        el_pos = self._get_electrodes()
        return pts, tri, el_pos

    def update(self, n_fan=8, n_layer=6, n_el=16):
        """update parameters"""
        self.n_fan = n_fan
        self.n_layer = n_layer
        self.n_el = n_el

    def _get_electrodes(self):
        """return the numbering of electrodes"""
        el_start = self.index_per_layer[self.n_layer - 1]
        el_len = self.pts_per_layer[self.n_layer]

        # place electrodes uniformly on the boundary
        n = np.linspace(
            el_start, el_start + el_len, num=self.n_el, endpoint=False, dtype=np.int
        )

        # for FMMU, electrodes should be placed clockwise
        # with 1 on the right (x+)
        n = n[::-1]  # reverse the order of electrodes
        n = np.roll(n, 1)  # rotate the index by 1
        return n

    def _spawn_points(self):
        """generate points"""
        # init points
        p = [0, 0]

        # divide r uniformly axial
        delta_r = 1.0 / self.n_layer

        for i in range(1, self.n_layer + 1):
            # increment points per-layer by fans
            n = i * self.n_fan
            r = i * delta_r
            # generate points on a layer
            pts = r * self._points_on_circle(n, offset=i)
            p = np.vstack([p, pts])

        return p

    @staticmethod
    def _points_on_circle(n, offset=0, offset_enabled=False):
        """generate points on unit circle"""
        fan_angle = 2 * np.pi / n
        a = np.array([i * fan_angle for i in range(n)])
        if offset_enabled:
            a += offset * (fan_angle / 2.0)
        pts = np.array([np.cos(a), np.sin(a)]).T

        return pts

    def _spawn_elements(self):
        """connect points fan-by-fan using a fixed pattern"""

        # element connections
        e = []
        for i in range(self.n_layer):
            e_layer = self._connect_layer(i)
            e.append(e_layer)

        return np.vstack(e)

    def _connect_layer(self, i):
        """
        generate connections on the i-th layer using points
        on the i-th and (i-1)-the layers.

        Notes
        -----
        make sure the triangles are counter-clock-wise (CCW)
        """

        # points per layer (ppl) in current and previous layer
        ppl_now = self.pts_per_layer[i + 1]
        ppl_pre = self.pts_per_layer[i]

        # starting index of current and previous layer
        index_now = self.index_per_layer[i]
        index_pre = self.index_per_layer[i - 1]

        e = []
        # A circle is divided into multiple fans, we assume that
        # points per layer per fan (plpf) is monotonically increasing
        point_plpf = i + 1
        # The connectivity pattern is the same for every fan,
        # but is different regarding whether the outer point (outer_now) is
        # interior point or boundary point on the current layer of a fan.
        # When the outer point is on the boundary,
        # the inner points (inner_now) is shared by adjacent fans,
        # thus k is not increasing and inner_now is reused.
        k = 0
        for j in range(ppl_now):

            # circular numbering in current layer
            outer_now = index_now + j
            outer_next = index_now + (j + 1) % ppl_now

            # circular numbering in previous layer
            inner_now = index_pre + k
            inner_next = index_pre + (k + 1) % ppl_pre

            # every (ppl_now/n_fan) points
            mode = j % point_plpf
            if mode == 0:
                # outer points is on the boundary
                ei = [outer_now, outer_next, inner_now]
                e.append(ei)
            else:
                # outer points is the interior points of a fan
                ei = [inner_now, outer_now, inner_next]
                e.append(ei)
                ei = [outer_now, outer_next, inner_next]
                e.append(ei)
                # advance k when outer points is a interior point
                k += 1

        return e


def demo():
    """demo using unit_circle_mesh"""
    model = MeshCircle()
    p, e, el_pos = model.create()

    # the order of inner-most triangles
    print(e[[0, 1, 2]])

    _, ax = plt.subplots(figsize=(6, 6))
    ax.plot(p[:, 0], p[:, 1], "ro", markersize=5)
    for i in range(p.shape[0]):
        ax.text(p[i, 0], p[i, 1], str(i))
    ax.set_xlim([-1.2, 1.2])
    ax.set_ylim([-1.2, 1.2])
    ax.grid("on")

    _, ax = plt.subplots(figsize=(6, 6))
    ax.triplot(p[:, 0], p[:, 1], e)
    ax.plot(p[el_pos, 0], p[el_pos, 1], "ro")
    for i, el in enumerate(el_pos):
        ax.text(p[el, 0], p[el, 1], str(i + 1))
    ax.set_xlim([-1.2, 1.2])
    ax.set_ylim([-1.2, 1.2])
    ax.grid("on")

    plt.show()


if __name__ == "__main__":
    demo()
